unit btp11;

{$G+}

(*************************************************************************)
(* Beni Tracker v1.8 Player - Pascal Version v1.1 rev. 0, by Zatzen 2020 *)
(*************************************************************************)

(* Einfache Vorgehensweise Beispiel:                    *)
(*                                                      *)
(* if btplay_loadmod('music003.op2') <> 0 then halt(1); *)
(* btplay_start_int(0, 0);                              *)
(* readln;                                              *)
(* btplay_stopmod;                                      *)
(* btplay_free_mod;                                     *)

interface

var
  btplay_next_tick: boolean;
(* wird von der zustndigen ISR 50x/Sek. auf true gesetzt,
   kann dann selbst bis zum nchsten mal auf false gesetzt werden *)

  btplay_usedheapmem: longint;
(* Variable fr verwendeten Heap-Speicher *)

  btplay_pisfilesize, btplay_pcsfilesize: longint;
(* werden vom pis2pcs Converter bentigt *)

function btplay_loadmod(filenm: string): byte;
(* unkomprimierte PIS Datei laden *)
(* Aufruf: btplay_loadmod(...);
   oder: err := btplay_mod(...);
   um einen evtl. Zugriffsfehler abzufangen *)

function btplay_load_pcs(filenm: string): byte;
(* komprimierte PCS Datei laden *)
(* s.o. *)

function btplay_save_pcs(filenm: string): byte;
(* funktional wenn zuvor PIS mit btplay_loadmod geladen wurde *)

procedure btplay_free_mod;
(* Speicher von geladener Datei auf dem Heap wieder freigeben *)

procedure btplay_start_int(position, row: integer);
(* Abspielen "im Hintergrund" initialisieren,
   ISR fhrt alle Berechnungen und OPL Zugriffe aus
   auerdem wird 50x/Sekunde btplay_next_tick auf TRUE gesetzt *)

procedure btplay_start_tickflag(position, row: integer);
(* ISR lediglich beauftragen,
   50x/Sekunde btplay_next_tick auf TRUE zu setzen *)

procedure btplay_playtick;
(* Kern der Abspielroutine. Kann in Zusammenhang mit
   btplay_next_tick ausserhalb der ISR Routine aufgerufen werden *)

procedure btplay_jumpto(position, row: integer; immediate: boolean);
(* Springen nach Position/Row.
   Wenn Immediate = TRUE wird sofort gesprungen und nicht
   das Ende der gerade gespielten Zeile abgewartet. *)

procedure btplay_stopmod;
(* Beendet die Wiedergabe und setzt den Timer wieder zurck *)




implementation

var
  pcsver, tabsiz_lo, tabsiz_hi, bitpsiz_lo, bitpsiz_hi,
  length, nptns, ninstrs, biggestptn, biggestins: byte;
  { genau so definiert lassen }

  tabdatasize: word absolute tabsiz_lo;
  bitpatdatasize: word absolute bitpsiz_lo;

type byte_struc = array[0..65534] of byte;
type ins_struc = array[0..31, 0..10] of byte;

var
  alloc_heap: array[1..9] of word;
  n, i: word; o, eh, el: byte;
  edrow, readrow: integer;
  v: word;
  oct: byte;

  ptnmap_poi: pointer;
  insmap_poi: pointer;
  old_timer_vec_read: boolean;

  ins_ofs, ins_seg: word;
  ordr_ofs, ordr_seg: word;
  pat_tablebits_ofs, pat_tablebits_seg: word;
  pat_tablofs_ofs, pat_tablofs_seg: word;
  pat_tables_ofs, pat_tables_seg: word;
  bitpat_ofs_ofs, bitpat_ofs_seg: word;
  bitpat_data_ofs, bitpat_data_seg: word;

  ins_poi: pointer absolute ins_ofs;
  ordr_poi: pointer absolute ordr_ofs;
  pat_tablebits_poi: pointer absolute pat_tablebits_ofs;
  pat_tablofs_poi: pointer absolute pat_tablofs_ofs;
  pat_tables_poi: pointer absolute pat_tables_ofs;
  bitpat_ofs_poi: pointer absolute bitpat_ofs_ofs;
  bitpat_data_poi: pointer absolute bitpat_data_ofs;
  ins: ^ins_struc absolute ins_poi;


  jump_override: boolean;
  count, speed: byte;
  edpos: word;
  e6xrow, e6xcount, e6x: integer;

  vocini: array[0..8] of shortint;
  vocnte, vocvol, vococt: array[0..8] of byte;

  vocfrq: array[0..8] of word;

  prtsrcfrq, prtdstfrq: array[0..8] of integer;
  prtsrco, prtdsto: array[0..8] of byte;

  arpi: word;
  flgarp: array[0..8] of byte;
  frqadd, prtadd: array[0..8] of integer;
  prtsign: array[0..8] of shortint;

  lastel, lasteh: array[0..8] of integer;

  arpfrq: array[0..2, 0..8] of word;
  arpoct: array[0..2, 0..8] of byte;

  posjmp, ptnbrk: integer;

  readpattern: word;

  old_intvec: pointer;
  tickervalue, tickerwrap: word;

const
  frqtab: array[0..11] of word =
  ($157, $16B, $181, $198, $1B0, $1CA, $1E5, $202, $220, $241, $263, $287);
  vocoff: array[0..8] of byte = (0, 1, 2, 8, 9, 10, 16, 17, 18);
  insofstab: array[0..31] of word =
  (0, 11, 22, 33, 44, 55, 66, 77, 88, 99, 110, 121,
   132, 143, 154, 165, 176, 187, 198, 209, 220, 231,
   242, 253, 264, 275, 286, 297, 308, 319, 330, 341);
  silentins: array[0..10] of byte = (0,0,$3F,$3F,$0F,$0F,$0F,$0F,0,0,0);




procedure read_ptn_row; assembler;
asm
  {
    - pat_tablebits bestimmen (bitbreite) (byte)
    - bitpat_ofs bestimmen (word)
    - bitpat-wert lesen (pattern, row)
    - pat_tablofs wert bestimmen (word)
    - an diesem offset + bitpat-wert * 3 drei Byte lesen
  }

  mov bx, readpattern
  mov es, pat_tablebits_seg
  mov si, pat_tablebits_ofs
  mov bl, es:[si+bx] { bx: bitbreite (bh = 0) }
  mov cx, bx
  add bx, bx
  mov bx, cs:[offset @jumptable + bx]
  xor ax, ax
  mov dx, readrow
  jmp bx
  @jumptable:
    dw offset @0bit
    dw offset @1bit
    dw offset @2bit
    dw offset @3bit
    dw offset @4bit
    dw offset @5bit
    dw offset @6bit

    @6bit: add ax, dx
    @5bit: add ax, dx
    @4bit: add ax, dx
    @3bit: add ax, dx
    @2bit: add ax, dx
    @1bit: add ax, dx
    @0bit:

  { ax: bitoffset ohne patternoffset }

  mov es, bitpat_ofs_seg
  mov si, bitpat_ofs_ofs
  mov bx, readpattern
  add bx, bx
  mov bx, es:[si+bx] { BX: Byte-offset im Bitpattern fr Row 0 }

  mov dx, ax
  shr dx, 3
  add bx, dx         { BX: Byte-offset im Bitpattern fr jeweilige Row }
                     {     ohne Bit-Drift }


  mov es, bitpat_data_seg
  mov si, bitpat_data_ofs

  mov bx, es:[si+bx]
  mov ch, cl { bits }
  mov cl, al
  and cl, 7 { bit-drift }
  shr bx, cl

  mov ax, 1
  mov cl, ch
  shl ax, cl
  dec ax      { maske }

  and bx, ax  { bx: gesuchte Position im Table }

  mov ax, bx
  add bx, bx
  add bx, ax  { * 3 }

  mov ax, readpattern
  add ax, ax
  mov es, pat_tablofs_seg
  mov si, pat_tablofs_ofs
  add si, ax
  mov ax, es:[si]   { ax: table-offset fr jeweiliges Pattern }

  mov es, pat_tables_seg
  mov si, pat_tables_ofs
  add si, ax
  mov ax, es:[si+bx]
  mov dl, es:[si+bx+2]

  mov el, dl
  mov dx, ax
  shr dl, 4
  mov eh, dl
  shr dh, 5
  mov o, dh
  mov dx, ax
  and dx, 15
  mov n, dx
  shr ax, 8
  and ax, 31
  mov i, ax

end;


procedure _opl_out_; assembler; { cl: reg, ch: vlu }
asm
  mov al, cl { reg }
  mov dx, 0388h
  out dx, al
  in al, dx
  in al, dx
  in al, dx
  in al, dx
  in al, dx
  in al, dx
  mov al, ch { vlu }
  inc dx
  out dx, al
  dec dx
  mov ch, 36
  @loop:
    in al, dx
    dec ch
    jnz @loop
end;

procedure _opl_setinstr_; assembler; { BX: Instrument }
asm
  mov cl, ds:[offset vocoff + di]
  add cl, 020h

  test bx, 32
  jz @normal
    mov ax, ds
    mov es, ax
    lea si, silentins
    xor bx, bx
    jmp @set
  @normal:
  mov es, ins_seg
  mov si, ins_ofs

  add bx, bx
  mov bx, ds:[offset insofstab + bx]

  @set:

  mov ch, es:[si+bx] {i.mul1}
  call _opl_out_

  add cl, 3
  mov ch, es:[si+bx+1] {i.mul2}
  call _opl_out_

  add cl, 01dh
  mov ch, es:[si+bx+2] {i.lev1}
  call _opl_out_

  add cl, 3
  mov ch, es:[si+bx+3] {i.lev2}
  call _opl_out_

  add cl, 01dh
  mov ch, es:[si+bx+4] {i.atd1}
  call _opl_out_

  add cl, 3
  mov ch, es:[si+bx+5] {i.atd2}
  call _opl_out_

  add cl, 01dh
  mov ch, es:[si+bx+6] {i.sur1}
  call _opl_out_

  add cl, 3
  mov ch, es:[si+bx+7] {i.sur2}
  call _opl_out_

  add cl, 05dh
  mov ch, es:[si+bx+8] {i.wav1}
  call _opl_out_

  add cl, 3
  mov ch, es:[si+bx+9] {i.wav2}
  call _opl_out_

  mov cx, v
  add cl, 0c0h
  mov ch, es:[si+bx+10] {i.fbcon}
  call _opl_out_

end;



procedure _calc_and_set_vol_; assembler;
{BX/BL: Instrument, DL: Parameter(Factor), DH: Basis}
{ lx := DH - (DL * (64 - ins[BX].levx)) shr 6; }
asm
  xor bh, bh
  add bx, bx
  mov bx, ds:[offset insofstab + bx]
  mov es, ins_seg
  mov si, ins_ofs
  mov bx, es:[si+bx+2]

  neg bl
  add bl, 64
  mov al, dl
  mul bl
  shr ax, 6
  neg al
  add al, dh
  mov bl, al

  neg bh
  add bh, 64
  mov al, dl
  mul bh
  shr ax, 6
  neg al
  add al, dh
  mov bh, al

  mov cl, ds:[offset vocoff + di]
  add cl, 040h
  mov ch, bl
  call _opl_out_
  add cl, 3
  mov ch, bh
  call _opl_out_

end;



procedure player_part_if_eh_3; assembler;
asm

mov bx, i
test bx, bx { if boolean(i) }
jz @skip3

  mov al, ds:[offset vocini + di]
  cmp bl, al { if vocini[v] <> i }
  je @skip1

    mov ds:[offset vocini + di], bl { di: v, bl: i }
    call _opl_setinstr_

  @skip1: { if vocini[v] <> i  end }

  mov al, ds:[offset vocvol + di]

  cmp al, 63 { if vocvol[v] < 63 }
  jae @skip2

    mov bx, i
    add bx, bx
    mov bx, ds:[offset insofstab + bx]
    mov es, ins_seg
    mov si, ins_ofs
    mov bx, es:[si+bx+2] { lev1 in bl, lev2 in bh }

    { lx := 64 - (64 * (64 - ins^[i].levx)) shr 6 = 64 - (64 - ins^[i].levx) = 64 - 64 + ins... = ins... }

    mov cl, ds:[offset vocoff + di]
    add cl, 040h
    mov ch, bl
    call _opl_out_
    add cl, 3
    mov ch, bh
    call _opl_out_

  @skip2: { if vocvol[v] < 63 end }
  mov al, 63
  mov ds:[offset vocvol + di], al

@skip3: {if boolean(i) end }

cmp n, 12 { if n < 12 }
jae @skip7

  mov al, ds:[offset vococt + di]
  mov oct, al { oct := vococt[v] }

  mov ds:[offset prtsrco + di], al { prtsrco[v] := oct }

  mov al, o
  mov ds:[offset prtdsto + di], al { prtdsto[v] := o }

  mov cx, ds:[offset vocfrq + bp] { frq := vocfrq[v] }

  mov ds:[offset prtsrcfrq + bp], cx  { prtsrcfrq[v] = frq }

  mov si, n
  add si, si
  mov ax, ds:[offset frqtab + si]
  mov ds:[offset prtdstfrq + bp], ax { prtdstfrq[v] := frqtab[n] }


  mov al, ds:[offset prtdsto + di]
  mov ah, oct
  mov dl, 1
  cmp al, ah { if prtdsto[v] < oct }
  je @skip5  { if not < or > oct }
  ja @skip4  { if > oct }

    neg dl

  @skip4:

    mov ds:[offset prtsign + di], dl  { prtsign[v] := (DL) }

    ret

  @skip5:

     mov ax, ds:[offset prtdstfrq + bp]
     {mov cx, frq} { CX enthlt bereits frq }
     cmp ax, cx { if prtdstfrq < frq }
     jae @skip6

       neg dl

     @skip6:
        mov ds:[offset prtsign + di], dl

@skip7:

end;


procedure player_part_if_boolean_i; assembler;
asm

cmp n, 12 (* if n < 12 *)
jae @skip6 (* sonst nach else springen *)

  mov cx, di
  add cl, 0b0h
  call _opl_out_ (* opl_out($B0 + v, 0) *)
  mov ax, -1
  mov ds:[offset lastel + bp], ax (* lastel[v] := -1 *)


  cmp eh, 0ch (* if eh <> $C *)
  je @skip3 (* sonst nach else springen *)


    mov bx, i
    mov al, ds:[offset vocini + di]
    cmp bl, al (* if vocini[v] <> i *)
    je @skip1

      mov ds:[offset vocini + di], bl (* vocini[v] := i *)
      (* in BX wird _opl_setinstr_ die Instrumentnummer bergeben *)
      call _opl_setinstr_

      jmp @skip2  (* else berspringen *)

    @skip1: (* else von if vocini[v] <> i *)


      mov al, ds:[offset vocvol + di]
      cmp al, 63 (* if vocvol[v] < 63 *)
      jae @skip2 (* sonst berspringen *)

        mov bx, i
        add bx, bx
        mov bx, ds:[offset insofstab + bx]
        mov es, ins_seg
        mov si, ins_ofs
        mov bx, es:[si+bx+2]

        mov cl, ds:[offset vocoff + di]
        add cl, 040h
        mov ch, bl
        call _opl_out_
        add cl, 3
        mov ch, bh
        call _opl_out_

    @skip2: (* if vocvol[v] < 63 ende *)
            (* if vocini[v] <> i else ende *)

      mov al, 63
      mov ds:[offset vocvol + di], al (* vocvol[v] := 63 *)

      jmp @skip5 (* else berspringen *)

  @skip3: (* if eh <> $C else *)


    mov bx, i
    mov al, ds:[offset vocini + di]
    cmp bl, al (* if vocini[v] <> i *)
    je @skip4  (* sonst berspringen *)

      mov ds:[offset vocini + di], bl
      call _opl_setinstr_ (* BX: i *)

    @skip4: (* if vocini[v] <> i ende *)

        mov bx, i
        mov dl, el
        mov dh, 62
        mov ds:[offset vocvol + di], dl
        call _calc_and_set_vol_


    @skip5: (* if eh <> $C else ende *)

      mov bx, n
      mov ds:[offset vocnte + di], bl
      add bx, bx
      mov bx, ds:[offset frqtab + bx]
      mov ds:[offset vocfrq + bp], bx { frq }

      mov cx, di
      add cl, 0a0h
      mov ch, bl
      call _opl_out_

      add cl, 010h
      mov ch, o
      mov ds:[offset vococt + di], ch
      shl ch, 2
      or ch, 32
      or ch, bh
      call _opl_out_

      ret (* vor Eintritt in  if n < 12 else zurckkehren *)


      { OK }

@skip6: (* if n < 12 else *)


  mov bx, i
  mov al, ds:[offset vocini + di]
  cmp bl, al (* if vocini[v] <> i *)
  je @skip8 (* sonst berspringen *)

    mov ds:[offset vocini + di], bl
    call _opl_setinstr_

    cmp eh, 0ch (* if eh = $C *)
    jne @skip7 (* sonst nach else springen *)

      mov bx, i
      mov dl, el
      mov dh, 62
      call _calc_and_set_vol_

      jmp @skip8 (* else berspringen *)


     @skip7: (* if eh = $C else *)

       mov al, ds:[offset vocvol + di]
       cmp al, 63 (* if vocvol[v] < 63 *)
       jae @skip8 (* sonst berspringen *)

        mov bx, i
        mov dl, ds:[offset vocvol + di]
        mov dh, 62
        call _calc_and_set_vol_


  @skip8:


    mov cx, ds:[offset lastel + bp] (* if boolean(lastel[v] *)
    jcxz @skip9 (* sonst berspringen *)


      mov ax, ds:[offset lasteh + bp]
      test ax, ax (* if lasteh[v] = 0 *)
      jnz @skip9 (* sonst berspringen *)

      mov bx, ds:[offset vocfrq + bp] { frq }
      mov cx, di
      add cl, 0a0h
      mov ch, bl
      call _opl_out_

      add cl, 010h
      mov ch, ds:[offset vococt + di]
      shl ch, 2
      or ch, 32
      or ch, bh
      call _opl_out_

@skip9:

end;


procedure player_part_if_boolean_i_else; assembler;
asm

  cmp n, 12 (* if n < 12 *)
  jae @skip_n_12_else (* sonst nach else springen *)

    mov ax, -1
    mov ds:[offset lastel + bp], ax (* lastel[v] := -1 *)

    cmp eh, 0ch (* if eh = $C *)
    jne @skip_if_eh_else (* sonst nach else springen *)

      mov bl, ds:[offset vocini + di]
      test bl, bl (* kleiner Null? *) (* if ti >= 0 *)
      js @skip_if_eh_else_end (* berspringen wenn < 0 *)

        mov dl, el
        mov dh, 62
        mov ds:[offset vocvol + di], dl
        call _calc_and_set_vol_

      jmp @skip_if_eh_else_end (* if eh ... else berspringen *)


    @skip_if_eh_else:

      mov al, ds:[offset vocvol + di]
      cmp al, 63 (* if vocvol[v] < 63 *)
      jae @skip_if_eh_else_end (* sonst berspringen *)

        mov bl, ds:[offset vocini + di]
        test bl, bl (* kleiner 0 ?   if ti >= 0 *)
        js @skip_if_eh_else_end (* wenn kleiner 0 berspringen *)

          mov dl, ds:[offset vocvol + di]
          mov dh, 64
          call _calc_and_set_vol_


    @skip_if_eh_else_end:

    mov bx, n
    add bx, bx
    mov bx, ds:[offset frqtab + bx] { frq }

    mov cx, di
    add cl, 0a0h
    mov ch, bl
    call _opl_out_

    add cl, 010h
    mov ch, o
    shl ch, 2
    or ch, 32
    or ch, bh
    call _opl_out_

    mov ax, n
    mov ds:[offset vocnte + di], al
    mov al, o
    mov ds:[offset vococt + di], al
    mov ax, bx
    mov bx, di
    mov ds:[offset vocfrq + bp], ax

    ret

  @skip_n_12_else:

    cmp eh, 0ch (* if eh = $C *)
    jne @skip_eh_0ch (* sonst berspringen *)

      mov bl, ds:[offset vocini + di]
      test bl, 128 (* eh >= 0 ? *)
      jnz @skip_eh_0ch (* sonst berspringen *)

        mov dl, el
        mov dh, 62
        mov ds:[offset vocvol + di], dl
        call _calc_and_set_vol_


    @skip_eh_0ch:

    mov cx, ds:[offset lastel + bp] (* if boolean(lasteh[v]) *)
    jcxz @skip_if_bool_lastel (* sonst berspringen *)

      mov ax, ds:[offset lasteh + bp]
      test ax, ax (* if lasteh[v] = 0 *)
      jnz @skip_if_bool_lastel (* sonst berspringen *)


        mov bx, ds:[offset vocfrq + bp]

        mov cx, di
        add cl, 0a0h
        mov ch, bl
        call _opl_out_
        mov ch, ds:[offset vococt + di]
        shl ch, 2
        or ch, 32
        or ch, bh
        add cl, 010h
        call _opl_out_



    @skip_if_bool_lastel:

end;


procedure player_part_fx_2; assembler;
asm
xor bx, bx
mov bl, el
shr bl, 4
sub bl, 6
js @end
and bl, 7
add bx, bx
mov bx, cs:[offset @jumptable + bx]

jmp bx

@jumptable:
dw offset @e6_
dw offset @end
dw offset @end
dw offset @end
dw offset @ea_
dw offset @eb_
dw offset @end
dw offset @end

@e6_:

  mov ax, e6x
  test ax, ax
  jnz @skip2 (* if e6x = 0 *)

    mov al, el
    test al, 15 (* if el and 15 = 0 *)
    jnz @skip1

      mov ax, edrow
      mov e6xrow, ax

      jmp @skip2

    @skip1: (* else *)

      mov al, el
      and ax, 15
      mov e6xcount, ax
      mov e6x, 1

  @skip2:

    mov cx, e6x (* if e6x = 1 *)
    jcxz @end

      mov al, el
      test al, 15 (* if el and 15 > 0 *)
      jz @end

        dec e6xcount
        mov ax, e6xcount
        test ax, ax (* if e6xcount >= 0 *)
        js @skip4

          mov ax, e6xrow
          dec ax
          mov edrow, ax

          jmp @end (* else berspringen *)

        @skip4: (* else *)

          mov e6x, 0

        jmp @end

@ea_:

  mov bl, ds:[offset vocini + di]
  test bl, bl (* if ti:= vocini[v] >= 0 *)
  js @end

    mov al, ds:[offset vocvol + di]
    mov dl, el
    and dl, 15
    add dl, al
    test dl, 011000000b
    jz @skp1
    mov dl, 63
    @skp1:
    mov dh, 64
    mov ds:[offset vocvol + di], dl

    call _calc_and_set_vol_


@eb_:

  mov bl, ds:[offset vocini + di]
  test bl, bl (* if ti:= vocini[v] >= 0 *)
  js @end

    mov dl, ds:[offset vocvol + di]
    mov al, el
    and al, 15
    sub dl, al
    cmp dl, 2
    jae @skp2
    mov dl, 2
    @skp2:
    mov dh, 64
    mov ds:[offset vocvol + di], dl
    call _calc_and_set_vol_




@end:

xor ax, ax
mov ds:[offset frqadd + bp], ax
mov ds:[offset prtadd + bp], ax
mov ds:[offset flgarp + di], al

end;


procedure player_part_fx; assembler;
asm

  xor bx, bx
  mov bl, eh
  add bx, bx
  mov cx, cs:[offset @jumptable + bx]
  xor ax, ax

  jmp cx

@jumptable:
  dw offset @eh_0
  dw offset @eh_1
  dw offset @eh_2
  dw offset @eh_3
  dw offset @end
  dw offset @end
  dw offset @end
  dw offset @end
  dw offset @end
  dw offset @end
  dw offset @end
  dw offset @eh_b
  dw offset @end
  dw offset @eh_d
  dw offset @eh_e
  dw offset @eh_f



@eh_0:

  mov al, el
  mov ah, ds:[offset lastel + bp]
  test al, al (* if boolean(el) *)
  jz @skip1 (* sonst berspringen *)

    cmp al, ah (* if el <> lastel[v] *)
    je @skip2 (* sonst berspringen *)

      xor ax, ax
      mov al, ds:[offset vocnte + di]
      mov n, ax
      mov al, ds:[offset vococt + di]
      mov o, al

      mov si, n
      mov cx, si
      add si, si
      mov ax, ds:[offset frqtab + si]
      mov ds:[offset arpfrq + bp], ax (* arpfrq[0, v] := frqtab[n] *)
      mov al, o
      mov ds:[offset arpoct + di], al
      mov dl, el
      mov dh, dl
      and dh, 15
      add dh, cl
      shr dl, 4
      add dl, cl

      (* an1 := n + el shr 4 *)
      (* an2 := n + el and 15 *)

      cmp dl, 12
      jae @skip3

        xor bx, bx
        mov bl, dl
        add bx, bx
        mov ax, ds:[offset frqtab + bx]
        mov ds:[offset arpfrq + bp + 18], ax
        (* arpfrq[1, v] := frqtab[an1] *)

        mov al, o
        mov ds:[offset arpoct + di + 9], al
        (* arpoct[1, v] := o *)

        jmp @skip4

      @skip3: (* else *)

        sub dl, 12
        xor bx, bx
        mov bl, dl
        add bx, bx
        mov dx, ds:[offset frqtab + bx]
        mov ds:[offset arpfrq + bp + 18], dx
        mov al, o
        inc al
        mov ds:[offset arpoct + di + 9], al

      @skip4: (* else end *)

      cmp dh, 12
      jae @skip5

        xor bx, bx
        mov bl, dh
        add bx, bx
        mov ax, ds:[offset frqtab + bx]
        mov ds:[offset arpfrq + bp + 36], ax
        mov al, o
        mov ds:[offset arpoct + di + 18], ax

        jmp @skip2

      @skip5: (* else *)

        xor bx, bx
        sub dh, 12
        mov bl, dh
        add bx, bx
        mov ax, ds:[offset frqtab + bx]
        mov ds:[offset arpfrq + bp + 36], ax
        mov al, o
        inc al
        mov ds:[offset arpoct + di + 18], al


    @skip2: (* if lastel end *)

      mov al, 1
      mov ds:[offset flgarp + di], al

      jmp @skip6

  @skip1: (* if boolean(el) else *)

    xor al, al
    mov ds:[offset flgarp + di], al

  @skip6:

    xor ax, ax
    mov ds:[offset frqadd + bp], ax
    mov ds:[offset prtadd + bp], ax

    ret




@eh_1:

  mov ds:[offset prtadd + bp], ax
  mov ds:[offset flgarp + di], al
  mov al, el
  mov ds:[offset frqadd + bp], ax

  ret



@eh_2:

  mov ds:[offset prtadd + bp], ax
  mov ds:[offset flgarp + di], al
  mov al, el
  neg ax
  mov ds:[offset frqadd + bp], ax

  ret



@eh_3:

  mov ds:[offset frqadd + bp], ax
  mov ds:[offset flgarp + di], al
  mov al, el
  mov ds:[offset prtadd + bp], ax

  ret



@eh_b:

  mov ds:[offset frqadd + bp], ax
  mov ds:[offset prtadd + bp], ax
  mov ds:[offset flgarp + di], al
  mov al, el
  mov posjmp, ax

  ret



@eh_d:

  mov ds:[offset frqadd + bp], ax
  mov ds:[offset prtadd + bp], ax
  mov ds:[offset flgarp + di], al
  mov al, el
  mov ptnbrk, ax

  ret



@eh_f:

  mov ds:[offset frqadd + bp], ax
  mov ds:[offset prtadd + bp], ax
  mov ds:[offset flgarp + di], al
  mov al, el
  mov speed, al

  ret



@eh_e:

  call player_part_fx_2


@end:

end;


procedure player_part_jumps; assembler;
asm

  mov al, jump_override
  test al, al
  jz @skip0
    mov jump_override, 0
    ret

  @skip0:
  mov ax, posjmp
  test ax, ax
  js @skip1

    mov edpos, ax

    mov ax, ptnbrk
    test ax, ax
    jns @skip2

      mov edrow, 0

      jmp @skip3

    @skip2: (* else *)

      mov edrow, ax
      mov ptnbrk, -1


    @skip3:

      mov posjmp, -1

      ret


  @skip1: (* else *)

    mov ax, ptnbrk
    test ax, ax
    js @skip4

      inc edpos
      mov dx, edpos
      mov dh, length
      xor dl, dh
      jnz @skip5
        mov edpos, 0
      @skip5:
      mov edrow, ax
      mov ptnbrk, -1

      ret


    @skip4: (* else *)

      inc edrow
      mov ax, edrow
      test ax, 64
      jz @skip6

        inc edpos
        mov ax, edpos
        mov ah, length
        xor al, ah
        jnz @skip7
          mov edpos, 0
        @skip7:
        mov edrow, 0


      @skip6:

end;


procedure player_part_arpi_prt; assembler;
asm
  mov cx, ds:[offset frqadd + bp]
  jcxz @skip1

    mov bx, ds:[offset vocfrq + bp]
    add bx, cx
    mov ds:[offset vocfrq + bp], bx

    mov cx, di
    add cl, 0a0h
    mov ch, bl
    call _opl_out_
    add cl, 010h
    mov ch, ds:[offset vococt + di]
    shl ch, 2
    or ch, 32
    or ch, bh
    call _opl_out_

    ret


  @skip1: (* else *)

    mov ax, ds:[offset prtadd + bp] (* hier kein JCXZ mglich da Sprungziel zu weit *)
    test ax, ax
    jz @skip2 (* if boolean(prtadd[v] *)

      mov bx, ds:[offset vocfrq + bp]
      mov al, ds:[offset vococt + di]
      mov oct, al


      mov al, ds:[offset prtsign + di] (* if prtsign[v] = 1 *)
      cmp al, 1
      jne @skip3

        mov ax, ds:[offset prtadd + bp]
        add bx, ax (* inc(frq, prtadd[v]) *)

        mov al, ds:[offset prtdsto + di]
        mov ah, oct
        xor al, ah
        jnz @skip4 (* if oct = prtdsto[v] *)

          mov ax, ds:[offset prtdstfrq + bp]
          cmp bx, ax
          jbe @skip4 (* if frq > prtdstfrq[v] *)

            (* frq = prtdstfrq[v] *)
            mov bx, ax

            xor ax, ax
            mov ds:[offset prtadd + bp], ax (* prtadd[v] := 0 *)





        @skip4:

          cmp bx, 0287h
          jbe @skip5 (* if frq > $287 *)

            sub bx, 0144h
            inc oct

            jmp @skip5



      @skip3: (* if prtsign[v] = 1 else *)



        mov ax, ds:[offset prtadd + bp]

        (* dec(frq, prtadd[v]) *)
        sub bx, ax


        mov al, ds:[offset prtdsto + di]
        mov ah, oct
        xor al, ah
        jnz @skip5 (* if oct = prtdsto[v] *)

          mov ax, ds:[offset prtdstfrq + bp]
          cmp bx, ax
          jae @skip6 (* if frq < prtdsto[v] *)

            (* frq := prtdstfrq[v] *)
            mov bx, ax

            xor ax, ax
            mov ds:[offset prtadd + bp], ax (* prtadd[v] := 0 *)

          @skip6:

            cmp bx, 0157h
            jae @skip5 (* if frq < $157 *)

              (* frq := $2AE - ($157 - frq *)
              add bx, 0157h
              dec oct



      @skip5: (* if prtsign[v] = 1 else end *)

        mov cx, di
        add cl, 0a0h
        mov ch, bl
        call _opl_out_

        add cl, 010h
        mov ch, oct
        shl ch, 2
        or ch, 32
        or ch, bh
        call _opl_out_

        mov ds:[offset vocfrq + bp], bx {frq}
        mov al, oct
        mov ds:[offset vococt + di], al

        ret


    @skip2: (* if boolean(prtadd[v]) else *)

      mov al, ds:[offset flgarp + di]
      test al, al
      jz @skip7

        mov ax, arpi
        mov si, ax
        shl si, 3
        add si, ax
        mov bx, si
        add si, si
        mov ax, ds:[offset arpfrq + bp + si]
        mov es, ax

        mov cx, di
        add cl, 0a0h
        mov ch, al
        call _opl_out_

        add cl, 010h

        mov ch, ds:[offset arpoct + di + bx]
        shl ch, 2
        or ch, 32
        mov ax, es
        or ch, ah
        call _opl_out_


      @skip7:

end;



procedure set_readpattern; assembler;
asm
  mov es, ordr_seg
  mov si, ordr_ofs

  mov bx, edpos
  mov ax, bx
  shl bx, 3
  add bx, ax
  add bx, v

  mov al, es:[si+bx]
  mov readpattern, ax { ah ist 0 }
end;



procedure btplay_playtick; assembler; { Haupt-Player-Routine }
asm
  push bp

  inc count
  mov al, count
  mov ah, speed
  xor al, ah
  jnz @skip1

    mov ax, edrow
    mov readrow, ax

    mov v, 0
    @loop1:

      mov di, v
      mov bp, di
      add bp,bp

      call set_readpattern
      call read_ptn_row


      mov al, eh
      cmp al, 3
      jne @else1
        call player_part_if_eh_3
        jmp @skipelse
      @else1:
        mov cx, i
        jcxz @else2
          call player_part_if_boolean_i
          jmp @skipelse
        @else2:
          call player_part_if_boolean_i_else

      @skipelse:

      call player_part_fx

      xor ax, ax
      mov al, eh
      mov bx, di
      mov ds:[offset lasteh + bp], ax
      mov al, el
      mov ds:[offset lastel + bp], ax

    inc v
    cmp v, 9
    jb @loop1

    call player_part_jumps

    mov count, 0


    pop bp
    ret


  @skip1:


    inc arpi
    cmp arpi, 3
    jne @skp2
      mov arpi, 0
    @skp2:

    mov v, 0
    @loop2:
      mov di, v
      mov bp, di
      add bp, bp

      call player_part_arpi_prt

      inc v
      cmp v, 9
      jb @loop2

  pop bp
end;


procedure define_pat_tablelen(srcpatdata, workdata: pointer; pattern: byte);
  var
    patsrc: ^byte_struc absolute srcpatdata;
    table: ^byte_struc absolute workdata;
    tblsize: ^byte_struc absolute pat_tablebits_poi;
    tablelen: integer;
    row, tabpos: word;
    note, oct, instr, fx_h, fx_l: byte;
    packet, tabvalue: longint;
    in_table: boolean;
    off: word;
begin
  packet := 0; tabvalue := 0;
  fillchar(table^[0], 192, 0);
  tablelen := 0;

  for row := 0 to 63 do
  begin
    note := patsrc^[row * 3] shr 4;
    oct := (patsrc^[row * 3] shr 1) and 7;
    if note = 12 then oct := 0;
    instr := (patsrc^[row * 3] and 1) shl 4 + patsrc^[row * 3 + 1] shr 4;
    fx_h := patsrc^[row * 3 + 1] and 15;
    fx_l := patsrc^[row * 3 + 2];

    packet := longint(note) + longint(fx_h) shl 4
      + longint(instr) shl 8 + longint(oct) shl 13 + longint(fx_l) shl 16;

    if row = 0 then
    begin
      move(packet, table^[0], 3);
      tablelen := 1;
    end
    else
    begin

      in_table := false;
      for tabpos := 0 to tablelen -1 do
      begin
        move(table^[tabpos * 3], tabvalue, 3);
        if packet = tabvalue then
          in_table := true;
      end;

      if not in_table then
      begin
        move(packet, table^[tablelen * 3], 3);
        inc(tablelen);
      end;

    end;

  end;

  tblsize^[pattern] := tablelen;

end;


procedure write_bitpat_data(srcpatdata: pointer; pattern, bits: word);
  type word_struc = array[0..32766] of word;
  var
    note, oct, instr, fx_h, fx_l: byte;
    packet, tabvalue: longint;
    bitoffset: longint;
    bitp_ofs: ^word_struc absolute bitpat_ofs_poi;
    bitp_data: ^byte_struc absolute bitpat_data_poi;
    row: word;
    tblpos: integer;
    tblofs: ^word_struc absolute pat_tablofs_poi;
    pat_tbl: ^byte_struc absolute pat_tables_poi;
    patsrc: ^byte_struc absolute srcpatdata;
    bitbuf: word;
    bitdrift: word;
    mask: word;
begin
  if bits = 0 then exit;
  packet := 0; tabvalue := 0;
  bitoffset := 0;
  for row := 0 to 63 do
  begin
    tblpos := 0;

    tblpos := -1;
    repeat
      inc(tblpos);

      note := patsrc^[row * 3] shr 4;
      oct := (patsrc^[row * 3] shr 1) and 7;
      if note = 12 then oct := 0;
      instr := (patsrc^[row * 3] and 1) shl 4 + patsrc^[row * 3 + 1] shr 4;
      fx_h := patsrc^[row * 3 + 1] and 15;
      fx_l := patsrc^[row * 3 + 2];

      packet := longint(note) + longint(fx_h) shl 4
        + longint(instr) shl 8 + longint(oct) shl 13 + longint(fx_l) shl 16;

      move(pat_tbl^[tblofs^[pattern] + tblpos * 3], tabvalue, 3);
    until packet = tabvalue;
    { Tabellenposition bestimmt, steht in tblpos }

    bitdrift := bitoffset and 7;
    mask := (1 shl bits - 1) shl bitdrift;
    move(bitp_data^[bitp_ofs^[pattern]+bitoffset shr 3], bitbuf, 2);
    mask := not mask;
    bitbuf := bitbuf and mask + tblpos shl bitdrift;
    move(bitbuf, bitp_data^[bitp_ofs^[pattern]+bitoffset shr 3], 2);

    {hier}

    inc(bitoffset, bits);
  end;

end;

function btplay_loadmod(filenm: string): byte;
  type word_struc = array[0..32766] of word;
  var
    f: file;
    a: word;
    io: byte;
    file_patpos: longint;
    patofs, pofs: word;
    pat_tmp_poi, table_tmp_poi: pointer;
    ptnmap: ^byte_struc absolute ptnmap_poi;
    insmap: ^byte_struc absolute insmap_poi;
    tbllen: ^byte_struc absolute pat_tablebits_poi;
    tblofs: ^word_struc absolute pat_tablofs_poi;
    pat_tbl: ^byte_struc absolute pat_tables_poi;
    bitp_ofs: ^word_struc absolute bitpat_ofs_poi;
    tables_offs: word;
    bitp_offs: word;
    bits: word;
begin
  assign(f, filenm);

  {$i-}
  reset(f, 1);
  {$i+}
  io := ioresult;
  if boolean(io) then
  begin
    btplay_loadmod := io;
    exit;
  end
  else
    btplay_loadmod := 0;

  btplay_pisfilesize := filesize(f);
  if btplay_pisfilesize = 0 then
  begin
    close(f);
    btplay_loadmod := 255; { leere Datei }
    exit;
  end;


  { "Header" }
  blockread(f, length, 3);

  { Pattern / Instrument - Mapper }
  getmem(ptnmap_poi, nptns);
  getmem(insmap_poi, ninstrs);
  alloc_heap[1] := nptns;
  alloc_heap[2] := ninstrs;
  blockread(f, ptnmap_poi^, nptns);
  blockread(f, insmap_poi^, ninstrs);

  getmem(ordr_poi, word(length) * 9); { Sequencer auf Heap einrichten + laden }
  alloc_heap[3] := word(length) * 9;
  blockread(f, ordr_poi^, word(length) * 9);

  { Grte Pattern-Nummer aus dem Mapper beziehen }
  { und anhand dessen Speicher reservieren }

  biggestptn := 0; for a := 0 to nptns - 1 do
    if ptnmap^[a] > biggestptn then biggestptn := ptnmap^[a];

  getmem(pat_tablofs_poi, (biggestptn + 1) shl 1);
  fillchar(pat_tablofs_poi^, (biggestptn + 1) shl 1, 0);
  alloc_heap[4] := (biggestptn + 1) shl 1;

  getmem(pat_tablebits_poi, biggestptn + 1);
  fillchar(pat_tablebits_poi^, biggestptn + 1, 0);
  alloc_heap[5] := biggestptn + 1;

  getmem(bitpat_ofs_poi, (biggestptn + 1) shl 1);
  fillchar(bitpat_ofs_poi^, (biggestptn + 1) shl 1, 0);
  alloc_heap[6] := (biggestptn + 1) shl 1;

  file_patpos := filepos(f);
  getmem(pat_tmp_poi, 192);
  getmem(table_tmp_poi, 192);
  tables_offs := 0; bitp_offs := 0;
  for a := 0 to nptns - 1 do
  begin
    blockread(f, pat_tmp_poi^, 192);
    define_pat_tablelen(pat_tmp_poi, table_tmp_poi, ptnmap^[a]);
    tblofs^[ptnmap^[a]] := tables_offs * 3;
    inc(tables_offs, tbllen^[ptnmap^[a]]);
    bitp_ofs^[ptnmap^[a]] := bitp_offs;
    case tbllen^[ptnmap^[a]] of
           2: inc(bitp_offs, 8);
        3..4: inc(bitp_offs, 16);
        5..8: inc(bitp_offs, 24);
       9..16: inc(bitp_offs, 32);
      17..32: inc(bitp_offs, 40);
      33..64: inc(bitp_offs, 48);
    end;
  end;

  freemem(table_tmp_poi, 192);
  freemem(pat_tmp_poi, 192);

  getmem(pat_tables_poi, tables_offs * 3);
  alloc_heap[7] := tables_offs * 3;
  getmem(bitpat_data_poi, bitp_offs);
  alloc_heap[8] := bitp_offs;
  {
  Gren der Bitpatterns:
  1 Bit: 8 Byte
  2 Bit: 16 Byte
  3 Bit: 24 Byte
  4 Bit: 32 Byte
  5 Bit: 40 Byte
  6 Bit: 48 Byte
  }
  getmem(pat_tmp_poi, 192);
  getmem(table_tmp_poi, 192);
  seek(f, file_patpos);
  for a := 0 to nptns - 1 do
  begin
    blockread(f, pat_tmp_poi^, 192);
    define_pat_tablelen(pat_tmp_poi, table_tmp_poi, ptnmap^[a]);
    { diesmal Tabellendaten sichern }
    move(table_tmp_poi^, pat_tbl^[tblofs^[ptnmap^[a]]], tbllen^[ptnmap^[a]]*3);

    case tbllen^[ptnmap^[a]] of
           1: bits := 0;
           2: bits := 1;
        3..4: bits := 2;
        5..8: bits := 3;
       9..16: bits := 4;
      17..32: bits := 5;
      33..64: bits := 6;
    end;
    write_bitpat_data(pat_tmp_poi, ptnmap^[a], bits);
    tbllen^[ptnmap^[a]] := bits; { !!! tablelen - > tablebits !!! }
  end;
  freemem(table_tmp_poi, 192);
  freemem(pat_tmp_poi, 192);

  biggestins := 0;
  for a := 0 to ninstrs - 1 do if insmap^[a] > biggestins then
    biggestins := insmap^[a];

  getmem(ins_poi, (biggestins+1) * 11);
  alloc_heap[9] := (biggestins+1) * 11;

  for a := 0 to ninstrs - 1 do blockread(f, ins^[insmap^[a]], 11);

  close(f);

  tabdatasize := tables_offs * 3;
  bitpatdatasize := bitp_offs;

  speed := 6; edpos := 0; edrow := 0;

  btplay_usedheapmem := 0;
  for a := 1 to 9 do inc(btplay_usedheapmem, alloc_heap[a]);

end; { loadmod }


function btplay_save_pcs(filenm: string): byte;
  var f: file;
  io: byte;
begin
  assign(f, filenm);
  {$i-}
  rewrite(f, 1);
  {$i+}
  io := ioresult;
  if boolean(io) then
  begin
    btplay_save_pcs := io;
    exit;
  end
  else
    btplay_save_pcs := 0;

  pcsver := 2;

  blockwrite(f, pcsver, 10);
  blockwrite(f, pat_tablebits_poi^, biggestptn + 1);
  blockwrite(f, pat_tablofs_poi^, (biggestptn + 1) shl 1);
  blockwrite(f, pat_tables_poi^, tabdatasize);
  blockwrite(f, bitpat_ofs_poi^, (biggestptn + 1) shl 1);
  blockwrite(f, bitpat_data_poi^, bitpatdatasize);
  blockwrite(f, ins_poi^, (biggestins + 1) * 11);
  blockwrite(f, ordr_poi^, word(length) * 9);

  btplay_pcsfilesize := filesize(f);
  close(f);
end;


function btplay_load_pcs(filenm: string): byte;
  var f: file;
  io, a: byte;
begin
  assign(f, filenm);
  {$i-}
  reset(f, 1);
  {$i+}
  io := ioresult;
  if boolean(io) then
  begin
    btplay_load_pcs := io;
    exit;
  end
  else
    btplay_load_pcs := 0;

  btplay_pcsfilesize := filesize(f);
  if btplay_pcsfilesize = 0 then
  begin
    close(f);
    btplay_load_pcs := 255; { leere Datei }
    exit;
  end;

  blockread(f, pcsver, 10);
  if pcsver <> 2 then
  begin
    close(f);
    btplay_load_pcs := 222; { nicht untersttzte PCS Version }
    exit;
  end;

  alloc_heap[1] := 0; alloc_heap[2] := 0;

  getmem(ordr_poi, word(length) * 9);
  alloc_heap[3] := word(length) * 9;

  getmem(pat_tablofs_poi, (biggestptn + 1) shl 1);
  alloc_heap[4] := (biggestptn + 1) shl 1;

  getmem(pat_tablebits_poi, biggestptn + 1);
  alloc_heap[5] := biggestptn + 1;

  getmem(bitpat_ofs_poi, (biggestptn + 1) shl 1);
  alloc_heap[6] := (biggestptn + 1) shl 1;

  getmem(pat_tables_poi, tabdatasize);
  alloc_heap[7] := tabdatasize;

  getmem(bitpat_data_poi, bitpatdatasize);
  alloc_heap[8] := bitpatdatasize;

  getmem(ins_poi, (biggestins + 1) * 11);
  alloc_heap[9] := (biggestins + 1) * 11;


  blockread(f, pat_tablebits_poi^, biggestptn + 1);
  blockread(f, pat_tablofs_poi^, (biggestptn + 1) shl 1);
  blockread(f, pat_tables_poi^, tabdatasize);
  blockread(f, bitpat_ofs_poi^, (biggestptn + 1) shl 1);
  blockread(f, bitpat_data_poi^, bitpatdatasize);
  blockread(f, ins_poi^, (biggestins + 1) * 11);
  blockread(f, ordr_poi^, word(length) * 9);

  close(f);

  btplay_usedheapmem := 0;
  for a := 1 to 9 do inc(btplay_usedheapmem, alloc_heap[a]);

  speed := 6; edpos := 0; edrow := 0;

end;


procedure btplay_free_mod;
begin
  freemem(ins_poi, alloc_heap[9]);
  freemem(bitpat_data_poi, alloc_heap[8]);
  freemem(pat_tables_poi, alloc_heap[7]);
  freemem(bitpat_ofs_poi, alloc_heap[6]);
  freemem(pat_tablebits_poi, alloc_heap[5]);
  freemem(pat_tablofs_poi, alloc_heap[4]);
  freemem(ordr_poi, alloc_heap[3]);
  if alloc_heap[1] > 0 then { wenn Mapper geladen wurden }
  begin
    freemem(insmap_poi, alloc_heap[2]);
    freemem(ptnmap_poi, alloc_heap[1]);
  end;
end;


procedure settimerfrq (hz: longint);
  var cnt: longint;
begin
  asm
    mov dx, 043h
    mov al, 034h
    out dx, al
  end;
  if hz = 0 then
  asm
    mov dx, 040h
    xor al, al
    out dx, al
    out dx, al
  end
  else
  begin
    tickervalue := $1234DD div hz;
    asm
      db 66h; mov bx, word ptr tickervalue
      mov dx, 040h
      mov al, bl
      out dx, al
      mov al, bh
      out dx, al
    end;
  end;
end;


procedure btplay_jumpto(position, row: integer; immediate: boolean);
begin
  jump_override := true;
  edpos := position;
  edrow := row;
  e6x := 0;
  if immediate then count := speed - 1;
end;


procedure introut_playtick; interrupt; assembler;
asm
  mov btplay_next_tick, 1
  call btplay_playtick
  mov ax, tickervalue
  add tickerwrap, ax
  jc @call_oldint
    mov al, 020h
    out 020h, al
    jmp @end
  @call_oldint:
    call old_intvec
  @end:
end;

procedure introut_flagtick; interrupt; assembler;
asm
  mov btplay_next_tick, 1
  mov ax, tickervalue
  add tickerwrap, ax
  jc @call_oldint
    mov al, 020h
    out 020h, al
    jmp @end
  @call_oldint:
    call old_intvec
  @end:
end;


procedure get_old_timer_vec; assembler;
asm
  mov al, old_timer_vec_read
  (* Absicherung old_intvec nicht mit eigener Routine zu berschreiben *)
  test al, al
  jnz @end
  xor ax, ax; mov es, ax
  db 66h; mov ax, es:[32]
  db 66h; mov word ptr old_intvec, ax
  mov old_timer_vec_read, 1
  @end:
end;

procedure set_timer_vec(p: pointer); assembler;
asm
  xor ax, ax; mov es, ax
  db 66h; mov ax, word ptr p
  cli; db 66h; mov es:[32], ax; sti
end;


procedure btplay_start_int(position, row: integer);
begin
  edpos := position;
  edrow := row;
  for v := 0 to 8 do
    vocini[v] := -1;
  posjmp := -1;
  ptnbrk := -1;
  e6x := 0; { e6 }
  count := speed - 1;
  settimerfrq(50);
  get_old_timer_vec;
  set_timer_vec(@introut_playtick);
end;

procedure btplay_start_tickflag(position, row: integer);
begin
  edpos := position;
  edrow := row;
  for v := 0 to 8 do
    vocini[v] := -1;
  posjmp := -1;
  ptnbrk := -1;
  e6x := 0; { e6 }
  count := speed - 1;
  settimerfrq(50);
  get_old_timer_vec;
  set_timer_vec(@introut_flagtick);
end;


procedure btplay_stopmod;
begin
  settimerfrq(0);
  set_timer_vec(old_intvec);

  for v := 0 to 8 do
  asm
    mov di, v
    mov bx, 32
    call _opl_setinstr_
    mov al, 32
    mov ds:[offset vocini + di], al
    mov cx, di
    add cl, 0b0h
    call _opl_out_
  end;
end;


begin
  old_timer_vec_read := false; { read im Sinne von gelesen, nicht zu lesen! }
end.